@@ -54,7 +54,7 @@
 #include "rpc_com.h"
 #include "debug.h"
 
-#define	su_data(xprt)	((struct svc_dg_data *)(xprt->xp_p2))
+#define	su_data(xprt)	((struct svc_dg_data *)((xprt)->xp_p2))
 #define	rpc_buffer(xprt) ((xprt)->xp_p1)
 
 #ifndef MAX
@@ -96,6 +96,7 @@ svc_dg_create(fd, sendsize, recvsize)
 	u_int recvsize;
 {
 	SVCXPRT *xprt;
+	SVCXPRT_EXT *ext = NULL;
 	struct svc_dg_data *su = NULL;
 	struct __rpc_sockinfo si;
 	struct sockaddr_storage ss;
@@ -120,6 +121,11 @@ svc_dg_create(fd, sendsize, recvsize)
 		goto freedata;
 	memset(xprt, 0, sizeof (SVCXPRT));
 
+	ext = mem_alloc(sizeof (*ext));
+	if (ext == NULL)
+		goto freedata;
+	memset(ext, 0, sizeof (*ext));
+
 	su = mem_alloc(sizeof (*su));
 	if (su == NULL)
 		goto freedata;
@@ -131,7 +137,7 @@ svc_dg_create(fd, sendsize, recvsize)
 	su->su_cache = NULL;
 	xprt->xp_fd = fd;
 	xprt->xp_p2 = su;
-	xprt->xp_auth = NULL;
+	xprt->xp_p3 = ext;
 	xprt->xp_verf.oa_base = su->su_verfbody;
 	svc_dg_ops(xprt);
 	xprt->xp_rtaddr.maxlen = sizeof (struct sockaddr_storage);
@@ -151,6 +157,8 @@ freedata:
 	if (xprt) {
 		if (su)
 			(void) mem_free(su, sizeof (*su));
+		if (ext)
+			(void) mem_free(ext, sizeof (*ext));
 		(void) mem_free(xprt, sizeof (SVCXPRT));
 	}
 	return (NULL);
@@ -251,8 +259,9 @@ svc_dg_reply(xprt, msg)
 	XDR_SETPOS(xdrs, 0);
 	msg->rm_xid = su->su_xid;
 	if (xdr_replymsg(xdrs, msg) &&
-	    (!has_args || (xprt->xp_auth &&
-	     SVCAUTH_WRAP(xprt->xp_auth, xdrs, xdr_results, xdr_location)))) {
+	    (!has_args ||
+	     SVCAUTH_WRAP(&SVC_XP_AUTH(xprt),
+			  xdrs, xdr_results, xdr_location))) {
 		struct msghdr *msg = &su->su_msghdr;
 		struct iovec iov;
 
@@ -279,8 +288,9 @@ svc_dg_getargs(xprt, xdr_args, args_ptr)
 	xdrproc_t xdr_args;
 	void *args_ptr;
 {
-	if (! SVCAUTH_UNWRAP(xprt->xp_auth, &(su_data(xprt)->su_xdrs),
-			     xdr_args, args_ptr)) {
+	if (!SVCAUTH_UNWRAP(&SVC_XP_AUTH(xprt),
+			    &(su_data(xprt)->su_xdrs),
+			    xdr_args, args_ptr)) {
 		return FALSE;
 	}
 	return TRUE;
@@ -302,18 +312,16 @@ static void
 svc_dg_destroy(xprt)
 	SVCXPRT *xprt;
 {
+	SVCXPRT_EXT *ext = SVCEXT(xprt);
 	struct svc_dg_data *su = su_data(xprt);
 
 	xprt_unregister(xprt);
 	if (xprt->xp_fd != -1)
 		(void)close(xprt->xp_fd);
-	if (xprt->xp_auth != NULL) {
-		SVCAUTH_DESTROY(xprt->xp_auth);
-		xprt->xp_auth = NULL;
-	}
 	XDR_DESTROY(&(su->su_xdrs));
 	(void) mem_free(rpc_buffer(xprt), su->su_iosz);
 	(void) mem_free(su, sizeof (*su));
+	(void) mem_free(ext, sizeof (*ext));
 	if (xprt->xp_rtaddr.buf)
 		(void) mem_free(xprt->xp_rtaddr.buf, xprt->xp_rtaddr.maxlen);
 	if (xprt->xp_ltaddr.buf)
